/*PACK*/
// ================================================== //
//          XIMSS Client Library (JavaScript)         //
//                                                    //
// Version 1.0.26                                     //
// Copyright (c) 2010-2014, Stalker Software, Inc.    //
// ================================================== //

function XIMSSSession(params,openCallbackRef,openCallbackMethod) {
  var theSessionRef       = this;
  function debugLog(x,y) {theSessionRef.debugFunction("XIMSS",x,y);}

  var errorCode           = null;


  var MaxBufferedRequests = 10;
  var ximssRequestSeqId   = 0;
  var isDead = false,isStarted = false,
      isClosed = false,                                // "<bye>" has been sent
      delayResponseProcessing = false;

  var asyncRegistry       = new XIMSSListenerRegistry;
  var asyncUnknown        = null;

  var pendingRequests     = new Object;
  var collectedReqIds     = new Array;
  var sendCollectedNow    = false;

  var initialXMLs         = new Array;

  var theWindow           = window;                     // @1.0.19
  
  function scheduleDelayed(f,millisecs) {
    return(theWindow.setTimeout(f,millisecs));
  }
  function cancelDelayed(theTimer) {
    if(theTimer != null) theWindow.clearTimeout(theTimer);
  }

  function createCallback(theRef, theMethod) {
    return function () {return(theMethod.call(theRef));};
  }

  function createCallback1(theRef, theMethod) {
    return function(arg1) {return(theMethod.call(theRef, arg1));};
  }
  function createCallback2(theRef, theMethod) {
    return function(arg1,arg2) {return(theMethod.call(theRef, arg1, arg2));};
  }

  this.setDebugFunction   = function(newDebugObject,newDebugMethod) {
    this.debugMethod = newDebugMethod;
    this.debugObject = newDebugObject;
  };


  function canAssign(x,y) {return((x == null) != (y == null));}


  this.setAsyncProcessor = function(theRef,newCallback,theTagName,theAttrName,theAttrValue) {
    if(newCallback != null && theRef != null) newCallback = createCallback1(theRef,newCallback);
    return(asyncRegistry.registerCallback(newCallback,theTagName,theAttrName,theAttrValue));
  };

  this.setUnknownAsyncProcessor = function(theRef,newCallback) {
    if(newCallback != null && theRef != null) newCallback = createCallback1(theRef,newCallback);
    var retCode = canAssign(newCallback,asyncUnknown);
    if(retCode) {asyncUnknown = newCallback;}
    return(retCode);
  };


  var  netErrorCallback      = null,
       netOKCallback         = null,
       netErrorTimeReported  = new Date(),
       netErrorReported      = false,
       netErrorTimeLimit     = 0;

  this.setNetworkErrorProcessor = function(theRef,newCallback,timeLimit) {
    if(newCallback != null && theRef != null) newCallback = createCallback2(theRef,newCallback);
    var retCode = canAssign(newCallback,netErrorCallback);
    if(retCode) {netErrorCallback = newCallback; netErrorTimeLimit = (timeLimit == null ? 10 : timeLimit);}    
    return(true);
  };

  this.setNetworkOKProcessor = function(theRef,newCallback) {
    if(newCallback != null && theRef != null) newCallback = createCallback(theRef,newCallback);
    var retCode = canAssign(newCallback,netOKCallback);
    if(retCode) {netOKCallback = newCallback;}    
    return(true);
  };

  function reportNetworkError(isFatal,elapsed) {
    if(isDead) return(true);
    var thisTime = new Date();
    if(isFatal || (netErrorCallback != null && netErrorTimeReported.getTime() + netErrorTimeLimit*1000 < thisTime.getTime())) {
      if(netErrorCallback != null) {
        netErrorReported = true;
        isFatal |= netErrorCallback(isFatal,elapsed/1000);
        netErrorTimeReported = thisTime;
      }
    }
    if(isFatal) {interruptAllPending();}
    return(!isFatal);
  }

  function reportNetworkOK() {
    if(isDead) return;
    if(netOKCallback != null && netErrorReported) {
      netErrorReported = false;
      netOKCallback();
    }
  }
  
  var sendRequestAndUnlock = function() {throw "not implemented";};
  var startConnection      = function() {throw "not implemented";};
  var interruptConnection  = function() {throw "not implemented";};

  this.sendRequest = function(xmlRequest,theRef,dataCallback,finalCallback,sendImmediately) {
    if(theRef != null) {
      if(dataCallback  != null) dataCallback  = createCallback2(theRef,dataCallback);
      if(finalCallback != null) finalCallback = createCallback2(theRef,finalCallback);
    }
  
    ximssRequestSeqId %= 1000000;                         // limit to 6 digits, we hope there would be no coflict
    var theSeqId  = '' + (++ximssRequestSeqId);
    xmlRequest.setAttribute("id",theSeqId);

    // if the session is already dead, return an error immediately, and do nothing more
    if(isDead || !isStarted || isClosed) {
      if(finalCallback != null) finalCallback("communication error",xmlRequest);
      return;
    }

    debugLog("out:",xmlRequest);

    var theRequest = pendingRequests[theSeqId] = new Object;
    theRequest.xmlRequest    = xmlRequest;
    theRequest.dataCallback  = dataCallback;
    theRequest.finalCallback = finalCallback;
    if(xmlRequest.tagName == "bye") {isClosed = true; delayResponseProcessing = false;}  // @1.0.23: response to bye should be processed synchronously, so we will not re-start reading.

    collectedReqIds.push(theSeqId);     
    sendCollectedNow |= ((sendImmediately !== false) || collectedReqIds.length >= MaxBufferedRequests);

    sendRequestAndUnlock();
  };
  
  var responseQueue = [];
  var nScheduled    = 0;
  function processXMLResponseQueue() {
    --nScheduled;
    var xResponse;
    for(xResponse = 0; xResponse < 10 && responseQueue.length != 0; ++xResponse) {    // @1.0.21
      var xmlResponse = responseQueue.shift();
//    if(responseQueue.length != 0) scheduleDelayed(processXMLResponseQueue,0); // schedule before execution, as it can fail
      if(responseQueue.length != 0 && nScheduled == 0) {++nScheduled; scheduleDelayed(processXMLResponseQueue,0);}  // @1.0.21
      syncProcessXMLResponse(xmlResponse);
    }
  }

  function processXMLResponse(xmlResponse) {
    if(isDead) return;
    if(delayResponseProcessing) {
      debugLog("a-inp(" + responseQueue.length + "):",xmlResponse);
//    scheduleDelayed(function() {syncProcessXMLResponse(xmlResponse)},0);  1.0.19: cannot do that, as some browsers can run scheduled tasks in the wrong order
//    if(responseQueue.length == 0) scheduleDelayed(processXMLResponseQueue,0);
      if(nScheduled == 0) {++nScheduled; scheduleDelayed(processXMLResponseQueue,0);}  // @1.0.21
      responseQueue.push(xmlResponse);
    } else {
      syncProcessXMLResponse(xmlResponse);
    }
  }
 
  function syncProcessXMLResponse(xmlResponse) {
    if(isDead) return;
    debugLog("inp:",xmlResponse);
 
    var theSeqId  = xmlResponse.getAttribute("id");
    if(theSeqId != null) {
      if(theSeqId != "noop") {
        var theDescr = pendingRequests[theSeqId];

        if(theDescr == null) {
          debugLog("io: unexpected response. id=" + theSeqId);
        } else if(xmlResponse.tagName == "response") {
          delete pendingRequests[theSeqId];
          if(theDescr.xmlRequest.tagName == "bye") {debugLog("io: BYE accepted"); interruptAllPending();}
          if(theDescr.finalCallback != null) theDescr.finalCallback(xmlResponse.getAttribute("errorText"),theDescr.xmlRequest);
        } else {
          if(theDescr.dataCallback != null) theDescr.dataCallback(xmlResponse,theDescr.xmlRequest);
        }
      }

    } else if(xmlResponse.tagName == "bye") {
      debugLog("io: BYE received");
      reportNetworkError(true,0);

    } else {                                                         // no "id" attribute -> async message
      debugLog("async: " + xmlResponse.tagName);
      if(!asyncRegistry.callCallbacks(xmlResponse)) {
        if(asyncUnknown != null) asyncUnknown(xmlResponse);
      }
    }
  }

  this.setDelayedResponseProcessing = function(x) { delayResponseProcessing = x;};

  function clearCollectedRequestsUnderLock() {
    while(collectedReqIds.length != 0) collectedReqIds.pop();
    sendCollectedNow = false;
  }

  function interruptAllPending() {
    isDead = true;
    debugLog("io: session closed");
    clearCollectedRequestsUnderLock();

    for(id in pendingRequests) {
      theDescr = pendingRequests[id];
      if(theDescr.finalCallback != null) theDescr.finalCallback("communication error",theDescr.xmlRequest); 
    }
    pendingRequests = new Object;
    interruptConnection();
  }


  this.start = function() {
    if(isDead || isStarted) {return;}

    while(initialXMLs.length != 0) processXMLResponse(initialXMLs.shift());
    isStarted = true;
    startConnection();
  };

  this.close = function(theRef,finalCallback) {
    this.sendRequest(this.createXMLNode("bye"),theRef,null,finalCallback,true);
  };

  //
  // HTTP implementation
  //
  function HTTPBinding() {  
    var httpBaseURL;
    var httpPOSTTimeOut =10,asyncGetWaitInSeconds = 15;
    var useAsyncPost,useNoAsync;
  
    var sendBatchSeqId  = 0,readBatchAckId = 0;
    var POSTURLRandom   = 0,GETURLRandom   = 0;
    var POSTInProgress  = false;
    var POSTHTTPRequest = null,GETHTTPRequest = null;

    var currentPOSTURL  = null, currentGETURL = null;
    var currentPOSTBody = null;
    var currentPOSTErrors = 0,currentGETErrors = 0;
    var currentPOSTComposed, currentGETComposed;
    var currentPOSTStarted,  currentGETStarted;
    
    var POSTTimer       = null,GETTimer = null;
    
    var POSTXMLDocument = null;

    var sessionLost = "you have been disconnected";
    function isFatalError(errorCode) {
      return(errorCode == sessionLost);
    }

    function ioDebugLog(x,y) {theSessionRef.debugFunction("HTTP",x,y);}

    function composePOSTUnderLock() { 
      currentPOSTURL = httpBaseURL + (params.asyncMode == "asyncPOST" ? "async" : "sync") + "?reqSeq=" + (++sendBatchSeqId);
      if(params.asyncMode == "noAsync") {currentPOSTURL += "&syncAsync=1";}

      for(var x; (x = POSTXMLDocument.firstChild) != null;) POSTXMLDocument.removeChild(x);
      currentPOSTBody = POSTXMLDocument.createElement("XIMSS");
      ioDebugLog("POST composed for " + collectedReqIds.length + " requests");
      if(collectedReqIds.length > 0) {
        for(var i = 0; i < collectedReqIds.length; ++i) {
          currentPOSTBody.appendChild(pendingRequests[collectedReqIds[i]].xmlRequest); 
        }
      } else {
        var xmlNoop = POSTXMLDocument.createElement("noop");
        xmlNoop.setAttribute("id","noop");
        currentPOSTBody.appendChild(xmlNoop);
      }
      POSTXMLDocument.appendChild(currentPOSTBody);
      currentPOSTBody = POSTXMLDocument;

      clearCollectedRequestsUnderLock();

      currentPOSTErrors   = 0;
      currentPOSTComposed = netErrorTimeReported = new Date();

      POSTInProgress = true;
    }

    function startComposedPOST() {
      currentPOSTStarted = new Date();
      updatePOSTTimeout();
      POSTHTTPRequest    = theSessionRef.startHTTPRequest(processPOSTCompletion,updatePOSTTimeout,"POST",currentPOSTURL+"&random=" + (++POSTURLRandom),currentPOSTBody);
      ioDebugLog("POST sent");
    }
    
    function updatePOSTTimeout() {
      if(POSTTimer != null) {cancelDelayed(POSTTimer);}
      POSTTimer          = scheduleDelayed(processPOSTTimeout,httpPOSTTimeOut*1000);
    }

    function processPOSTTimeout() {
      POSTTimer = null;
      if(POSTHTTPRequest != null) {
        ioDebugLog("POST time-out");
        cancelHTTPRequest(POSTHTTPRequest);
        processPOSTCompletion(null,"request time-out");
      }
    }

    function startNoopPOST() {
      if(!POSTInProgress) {
        POSTTimer = null;
        if(!isClosed && !isDead) {                                        // @1.0.22
          composePOSTUnderLock();
          startComposedPOST();
        }
      }
    }


    function processPOSTCompletion(xmlData,errorCode) {
      ioDebugLog("POST completed" + (errorCode == null ? "" : " errorCode="+errorCode));
      if(POSTHTTPRequest == null) {ioDebugLog("POST missing"); return;}

      if(POSTTimer != null) {cancelDelayed(POSTTimer); POSTTimer = null;}
      POSTHTTPRequest = null;
      var isFatal = isFatalError(errorCode);

      if(errorCode == null && params.asyncMode != "asyncPOST") {
        if(xmlData == null) {
          errorCode = "no XML in POST response";
        } else {
          for(var i = 0; i < xmlData.childNodes.length; ++i) {
            processXMLResponse(xmlData.childNodes[i]);
          }
        }
      }

      if(isDead) {                                        // we have been closed/suspended
        POSTInProgress = false;
      } else if(errorCode == null) {
        reportNetworkOK();
        POSTInProgress = false;
        if(sendCollectedNow) {
          composePOSTUnderLock();
          startComposedPOST();
        } else if(params.pollPeriod != null) {
          POSTTimer = scheduleDelayed(startNoopPOST,params.pollPeriod*1000);
        }
      } else {
        if((isFatal || ++currentPOSTErrors >= 2) && !reportNetworkError(isFatal,(new Date()).getTime() - currentPOSTComposed.getTime())) {
          POSTInProgress = false;
          return;
        }

        var delayTime = 
          currentPOSTErrors >= 20 ? 10000 : 
          currentPOSTErrors >= 10 /*|| errorCode == "no network available" */? 5000 :
          currentPOSTErrors >= 5  /*|| errorCode == "no route to host"     */? 2000 : 
                                                                               200; 

        delayTime = currentPOSTStarted.getTime() + delayTime - (new Date()).getTime();
        if(delayTime < 100) delayTime = 100;
        ioDebugLog("POST paused for " + delayTime + " msecs");
        POSTTimer = scheduleDelayed(startComposedPOST,delayTime);
      }
    }

    sendRequestAndUnlock = function() {
      if(POSTInProgress || !sendCollectedNow) {
        ioDebugLog("POST postponed");
        return;
      }
      if(POSTTimer != null) {cancelDelayed(POSTTimer); POSTTimer  = null;}
      composePOSTUnderLock();
      startComposedPOST();
    };

    function startComposedGET() {
      currentGETStarted = new Date();
      updateGETTimeout();
      GETHTTPRequest    = theSessionRef.startHTTPRequest(processGETCompletion,updateGETTimeout,"GET",currentGETURL+"&random=" + (++GETURLRandom),null);
      ioDebugLog("GET sent");
    }

    function updateGETTimeout() {
      if(GETTimer != null) {cancelDelayed(GETTimer);}
      GETTimer  = scheduleDelayed(processGETTimeout,(asyncGetWaitInSeconds+2)*1000);
    }

    function processGETTimeout() {
      GETTimer = null;
      if(GETHTTPRequest != null) {
        ioDebugLog("GET time-out");
        cancelHTTPRequest(GETHTTPRequest);
        processGETCompletion(null,"request time-out");
      }
    }

    function composeAndStartGET() {
      currentGETURL = httpBaseURL + "?ackSeq="  + readBatchAckId + 
                                    "&maxWait=" + asyncGetWaitInSeconds;

      startComposedGET();
      currentGETComposed = netErrorTimeReported = new Date();
    }

    function processGETCompletion(xmlData,errorCode) {
      ioDebugLog("GET completed" + (errorCode == null ? "" : " errorCode="+errorCode) + (xmlData == null ? "" : " data read"));
      if(GETHTTPRequest == null) {ioDebugLog("GET missing"); return;}

      if(GETTimer != null) {cancelDelayed(GETTimer); GETTimer = null;}
      GETHTTPRequest    = null;

      var isFatal       = isFatalError(errorCode);

      if(errorCode != null) {
      } else if(xmlData == null) {
        errorCode = "no XML in GET response";
      } else {
        var respReqAttr = xmlData.getAttribute("respSeq");
        if(respReqAttr != null) readBatchAckId = respReqAttr;
        for(var i = 0; i < xmlData.childNodes.length; ++i) {
          processXMLResponse(xmlData.childNodes[i]);
        }
      }
   
      if(!isStarted || isDead) {                              // we have been closed
        ;
      } else if(errorCode == null) {
        reportNetworkOK();
        composeAndStartGET();
      } else {
        if((isFatal || ++currentGETErrors >= 2) && !reportNetworkError(isFatal,(new Date()).getTime() - currentGETComposed.getTime())) return;

        var delayTime = 
          currentGETErrors >= 20 && (params.asyncMode != "asyncPOST" || !POSTInProgress) ? 30000 : 
          currentGETErrors >= 10 /*|| errorCode == "no network available"*/ ? 5000 :
          currentGETErrors >= 5  /*|| errorCode == "no route to host"    */ ? 2000 : 
                                                                              200;
        delayTime = currentGETStarted.getTime() + delayTime - (new Date()).getTime();
        if(delayTime < 100) delayTime = 100;
        ioDebugLog("GET paused for " + delayTime + " msecs");
        GETTimer = scheduleDelayed(startComposedGET,delayTime);
      }
    }

    theSessionRef.startHTTPRequest = function(callback,timerUpdate,method,url,body,userName,userPassword) {
      var httpRequest = new XMLHttpRequest();
      if(httpRequest == null) {callback(null,"XMLHttpRequest failed");}

      httpRequest.onreadystatechange = function() {

        if(httpRequest.aborted) return;

        if(httpRequest.readyState != 4) {if(timerUpdate != null) {timerUpdate();} return;}

        var errorCode = null;
        var ximssData = null;
        if(httpRequest.status == 0) {
          errorCode = "server communication error";
        } else if(httpRequest.status == 550) {
          errorCode = sessionLost;
        } else if(httpRequest.status < 200 || httpRequest.status >= 300) {
          if((errorCode = httpRequest.statusText) == null) errorCode = "server protocol error";
        } else if(httpRequest.responseXML != null && (ximssData = httpRequest.responseXML.childNodes[0]) != null) { // @1.0.20
          if(ximssData.tagName != "XIMSS") {ximssData = null; errorCode = "server XML response is not XIMSS";}
          else if(POSTXMLDocument == null) POSTXMLDocument = httpRequest.responseXML;
        } else {
          if(httpRequest.responseText != null && httpRequest.responseText != "") errorCode = "server response is not XML";
        }
        callback(ximssData,errorCode);
      };

      httpRequest.onprogress = function() {
        if(timerUpdate != null && !httpRequest.aborted && httpRequest.readyState != 4) {timerUpdate();}             // @1.0.25: aborted, readyState check
      };

      httpRequest.open(method,url,true,userName,userPassword);
      if(typeof(body) == "string") httpRequest.setRequestHeader("Content-Type","text/xml; charset=utf-8");
      httpRequest.send(body);
      return(httpRequest);
    };

    function cancelHTTPRequest(httpRequest) {
      httpRequest.aborted = true;
      httpRequest.abort();
    }

    startConnection  = function() {
      if(params.asyncMode != "noAsync") composeAndStartGET();
    };

    interruptConnection = function() {
      if(POSTTimer != null) {cancelDelayed(POSTTimer); POSTTimer  = null;}
      if(GETTimer  != null) {cancelDelayed(GETTimer);  GETTimer   = null;}

      if(POSTHTTPRequest!= null) {ioDebugLog("interrupting POST"); cancelHTTPRequest(POSTHTTPRequest); POSTHTTPRequest = null;}
      if(GETHTTPRequest != null) {ioDebugLog("interrupting GET");  cancelHTTPRequest(GETHTTPRequest);  GETHTTPRequest  = null;}
    };

    theSessionRef.createXMLNode      = function(tagName) {
      if(POSTXMLDocument == null) throw "no POSTXMLDocument preserved";
      return(POSTXMLDocument.createElement(tagName));
    };
    theSessionRef.createTextNode     = function(text) {
      if(POSTXMLDocument == null) throw "no POSTXMLDocument preserved";
      return(POSTXMLDocument.createTextNode(text));
    };

    var loginURL;

    function loginHTTPCompletion(xmlData,errorCode) {

      if(errorCode == null && xmlData == null) errorCode = "empty response on login";
    
      if(errorCode == null) {
        var xmlFirst = xmlData.childNodes[0];
        if(xmlFirst != null && xmlFirst.tagName == "response") {
          errorCode = xmlFirst.getAttribute("errorText");
        } else if(xmlFirst != null && xmlFirst.tagName == "session") {
          var theURL = xmlFirst.getAttribute("urlID");
          if(theURL != null) {
            httpBaseURL += "Session/" + theURL + "/"; 
            for(var i = 0; i < xmlData.childNodes.length; ++i) {
              initialXMLs.push(xmlData.childNodes[i]);
            }
          } else {
            errorCode = "the 'session' 'urlID' data is missing";
          }
        } else {
          errorCode = "the 'session' data is missing";
        }
      }
      if(openCallbackMethod != null) openCallbackMethod.call(openCallbackRef,errorCode == null ? theSessionRef : null,errorCode);
    }


    function preSessionHTTPCompletion(xmlData,errorCode) {
      if(openCallbackMethod != null) openCallbackMethod.call(openCallbackRef,errorCode == null ? xmlData : null,errorCode);
    }

    function preLoginHTTPCompletion(xmlData,errorCode) {
      var nonce = null;
      if(errorCode == null && xmlData != null) {
        var xmlFirst = xmlData.childNodes[0];
        if(xmlFirst != null && xmlFirst.tagName == "features") {
          var canPlain   = false;
          for(var i = 0; i < xmlFirst.childNodes.length; ++i) {
            var xmlFeature = xmlFirst.childNodes[i];
            if(xmlFeature.tagName == "nonce") {
              nonce = xmlFeature.firstChild.nodeValue;
            } else if(xmlFeature.tagName == "sasl" && params.loginMethod == "auto") {
              if(xmlFeature.firstChild.nodeValue == "CRAM-MD5" && typeof CRYDigesterMD5 == "function") {
                params.loginMethod = "cram-md5";
              } else if(xmlFeature.firstChild.nodeValue == "PLAIN") {
                canPlain = true;
              }
            }
          }
          if(params.loginMethod == "auto" && canPlain) {params.loginMethod = "plain";}
        }
      }

      if(params["fixedIP"]   == "NO") {loginURL += "&DisableIPWatch=1";}
      if(params["useCookie"] == "NO") {loginURL += "&DisableUseCookie=1";}

      if(errorCode == null && typeof(params.userName) != "string") errorCode = "XIMSS 'userName' is not specified";

      var httpAuthUserName = null,httpAuthPassword = null;
      if(errorCode != null) {
        
      } else if(params.loginMethod == "sessionid") {
        if(typeof(params.sessionid) != "string") {
          errorCode = "XIMSS 'sessionid' is not specified";
        } else {
          loginURL  += "&userName="+encodeURIComponent(params.userName) + "&sessionid=" + encodeURIComponent(params.sessionid);
        }

      } else if(typeof(params.password) != "string") {
        errorCode = "XIMSS 'password' is not specified";

      } else if(params.loginMethod == "plain") {
        loginURL    += "&userName="+encodeURIComponent(params.userName) + "&password="+encodeURIComponent(params.password);

      } else if(params.loginMethod == "cram-md5") {
        if(typeof(nonce) != "string") {
          errorCode = "XIMSS 'nonce' is missing from the server prompt";
        } else {
          loginURL  += "&userName="+encodeURIComponent(params.userName) + "&nonce="+encodeURIComponent(nonce) + "&authData=" + (new CRYDigesterMD5()).cramBase64(params.password,nonce);
        }

      } else if(params.loginMethod == "auth") {
        httpAuthUserName = params.userName;
        httpAuthPassword = params.password;

      } else if(params.loginMethod == "auto") {
        errorCode = "no supported XIMSS login methods";

      } else {
        errorCode = "illegal XIMSS login method";
      }

      if(errorCode == null) {
        theSessionRef.startHTTPRequest(loginHTTPCompletion,null,"GET",loginURL,null,httpAuthUserName,httpAuthPassword);
      } else {
        loginHTTPCompletion(null,errorCode);
      }
      return(null);
    }

    // constructor

    if(typeof(params.serverName) != "string") return("illegal XIMSS http binding parameters");
    httpBaseURL  = "http";
    if(params.secureMode == "YES") {
      httpBaseURL  = "https";
      if(params.loginMethod == "auto") {params.loginMethod = "plain";}
    }
    httpBaseURL +=  "://" + params.serverName + "/";
    loginURL     = httpBaseURL + "XIMSSLogin/?errorAsXML=1";
    
    if(params.preSession != null) {
      theSessionRef.startHTTPRequest(preSessionHTTPCompletion,null,"POST",loginURL,"<XIMSS>" + params.preSession + "</XIMSS>");
    } else if(params.loginMethod == "auto" || params.loginMethod == "cram-md5") {
      theSessionRef.startHTTPRequest(preLoginHTTPCompletion,null,"POST",loginURL,"<XIMSS><listFeatures id=\"list\" /></XIMSS>");
    } else {
      preLoginHTTPCompletion(null,null);
    }
    return(null);
  }

 
  if(params.binding == "HTTP") {
    errorCode = HTTPBinding();
  } else {
    errorCode = "illegal binding";
  }
//if(errorCode != null && openCallbackMethod != null) openCallbackMethod.call(openCallbackRef,null,errorCode);
  if(errorCode != null && openCallbackMethod != null) scheduleDelayed(function() {openCallbackMethod.call(openCallbackRef,null,errorCode);},0); // @1.0.26
}

XIMSSSession.prototype.getVersion = function() {return("1.0.26");};

XIMSSSession.prototype.setCommonDebugFunction = function(newDebugObject,newDebugMethod) {
  XIMSSSession.prototype.debugMethod = newDebugMethod;
  XIMSSSession.prototype.debugObject = newDebugObject;
};

XIMSSSession.prototype.debugFunction = function() {
  if(this.debugMethod != null) this.debugMethod.apply(this.debugObject,arguments);
};

XIMSSSession.prototype.debugLog = function(x,y) {this.debugFunction("XIMSS",x,y);};

XIMSSSession.prototype.callback = function(theRef, theMethod) {
  return function () {return(theMethod.apply(theRef, arguments));}
};




//
// Callback Registry
//
function XIMSSListenerRegistry() {
  var anyTag   = null;
  var tags     = new Object;

  this.registerCallback = function(newCallback,theTagName,theAttrName,theAttrValue) {
    function canAssign(x,y) {return((x == null) != (y == null));}
 
    if(theTagName == null) {
      if(theAttrName != null || theAttrValue != null) {return(false);}
      if(!canAssign(newCallback,anyTag)) {return(false);}
      anyTag = newCallback;

    } else {
      var tagDescr = tags[theTagName];
      if(tagDescr == null) {
        if(newCallback == null) {return(false);}
        // creating a new tagDescr element
        tagDescr = tags[theTagName] = new Object;
        tagDescr.attrNames = null; tagDescr.anyAttr = null;
      }

      if(theAttrName != null) {
        var previous = null,attrNameDescr;
        for(;;) {
          attrNameDescr = previous == null ? tagDescr.attrNames : previous.next;
        if(attrNameDescr == null || attrNameDescr.name == theAttrName) break;
          previous = attrNameDescr;
        }
        if(attrNameDescr == null) {
          if(newCallback == null) {return(false);}
          attrNameDescr          = new Object;
          attrNameDescr.next     = null;
          attrNameDescr.name     = theAttrName;
          attrNameDescr.values   = new Object;
          attrNameDescr.nValues  = 0;
          attrNameDescr.anyValue = null;
          
          if(previous == null) tagDescr.attrNames = attrNameDescr;
          else                 previous.next      = attrNameDescr;
        }

        if(theAttrValue != null) {
          if(!canAssign(newCallback,attrNameDescr.values[theAttrValue])) {return(false);}
          if(newCallback != null) {attrNameDescr.values[theAttrValue] = newCallback; ++attrNameDescr.nValues;}
          else                    {delete attrNameDescr.values[theAttrValue]; --attrNameDescr.nValues;}
        } else {
          if(!canAssign(newCallback,attrNameDescr.anyValue)) {return(false);}
          attrNameDescr.anyValue = newCallback;
        }
        // if all callbacks for this attribute have been removed -> remove the attribute
        if(attrNameDescr.anyValue == null && attrNameDescr.nValues == 0) {
          if(previous == null) tagDescr.attrNames = attrNameDescr.next;
          else                 previous.next      = attrNameDescr.next;
        }
      } else {
        if(theAttrValue != null) return(false);                // if theAttrName == null, then theAttrValue must be null, too.
        if(!canAssign(newCallback,tagDescr.anyAttr)) {return(false);}
        tagDescr.anyAttr = newCallback;
      }
      // if all callbacks for this tag have been removed -> remove the tag
      if(tagDescr.anyAttr == null && tagDescr.attrNames == null) {
        delete tags[theTagName];
      }
    }
    return(true);
  };

  this.callCallbacks = function(xmlData) {
    var foundDescrs = new Array;

    var tagDescr = tags[xmlData.tagName];
    if(tagDescr != null) {
      for(var attrScan = tagDescr.attrNames;  attrScan != null; attrScan = attrScan.next) {
        var attrValue = xmlData.getAttribute(attrScan.name);
        if(attrValue != null) {
          var valueCallback = attrScan.values[attrValue];
          if(valueCallback     != null) foundDescrs.push(valueCallback);
          if(attrScan.anyValue != null) foundDescrs.push(attrScan.anyValue);
        }
      }
      if(tagDescr.anyAttr != null) foundDescrs.push(tagDescr.anyAttr);
    }
    if(anyTag != null) foundDescrs.push(anyTag);

    // now, process all found in the proper order.
    //  consider processed if retCode >= 0
    //  stop processing    if retCode >  0
    var processed = false;
    while(foundDescrs.length != 0) {
      var result = foundDescrs.shift()(xmlData);
      processed |= result >= 0;
    if(result > 0) break;
    }
    
    return(processed);
  };

}
